/*************************************************************************
* MediaCoder - The universal audio/video transcoder
*
* Copyright (c) 2005-06 Stanley Huang <reachme@citiz.net>
* All rights reserved.
*************************************************************************/

#include "config.h"
#include "processpil.h"
#include "datatypes.h"
#include "transcode.h"
#include "preference.h"
#include "containers.h"
#include "common.h"

static bool IsSBRAAC(T_ITEM* item)
{
	//FIXME: identify SBR AAC
	switch (item->target.audioformat) {
	case AC_NEROAAC:
	case AC_NEROREF:
	case AC_AACPLUS:
		return true;
	}
	return false;
}

bool CDefContainer::CheckAudioSupport(int id)
{
	if (mcinst.fAudioEncoding) return true;
	switch (id) {
	case AC_LAMEMP3:
	case AC_FAAC:
	case AC_PCM:
	case AC_FFMPEG:
	case AC_COPY:
	case AC_NONE:
		return true;
	default:
		return false;
	}
}

bool CAVIFormat::CheckAudioSupport(int id)
{
	switch (id) {
	case AC_LAMEMP3:
	case AC_FAAC:
	case AC_PCM:
	case AC_FFMPEG:
	case AC_COPY:
	case AC_NONE:
		return true;
	default:
		return false;
	}
}

char* CMatroska::UpdateSettings(T_ITEM* item, char* destfile, CONTAINER_TRACKS tracks[])
{
	CPrefBase prefbase(prefNodePath);
	PARAM_OPT opts[] = {
		{".options", ""},
		{0},
	};
	int i, count = 0;
	cmd.Format("%s -o \"%s\"",pref.GetValueStr(".path"),destfile);
	for (i=0; i<MAX_TRACKS; i++) {
		if (!tracks[i].filename) continue;
		if (tracks[i].delay) {
			cmd.AppendFormat(" --delay 0:%d", tracks[i].delay);
		}
		switch (tracks[i].type) {
		case CT_AUDIO:
			if (IsSBRAAC(item)) cmd += " --aac-is-sbr 0";
			cmd.AppendFormat(" -D -S \"%s\"", tracks[i].filename);
			count++;
			break;
		case CT_VIDEO:
			cmd.AppendFormat(" -A \"%s\"", tracks[i].filename);
			break;
		case CT_SUBTITLE:
			cmd.AppendFormat(" -s -A -D \"%s\"",tracks[i].filename);	
			break;
		default:
			continue;
		}
		count++;
	}
	if (item->target.videoformat == VC_H264)
		cmd += " --engage allow_avc_in_vfw_mode";
	pref.GenOptions(cmd, opts, ' ');
	return count == 0 ? 0 : cmd.GetBuffer();
}

char* CMatroska::GetVersionInfo()
{
	CPrefBase prefbase(prefNodePath);
	return GetEncoderVersionInfo(pref.GetValueStr(".path"), "-V", 0);
}

char* CMatroska::GetOutExt(T_ITEM* item)
{
	return (mcinst.fAudioEncoding) ? "mka" : "mkv";
}

int CMatroska::Mux()
{
	memset(&sp ,0, sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	return ShellExec(&sp);
}

int CMatroska::GetProgress()
{ 
	ShellWait(&sp, -1);
	ShellClean(&sp);
	return -1;
}

char* CMp4box::UpdateSettings(T_ITEM* item, char* destfile, CONTAINER_TRACKS tracks[])
{
	CPrefBase prefbase(prefNodePath);
	PARAM_OPT opts[] = {
		{".packed", "-packed"},
		{".keepSystemInfo", "-keepsys"},
		{".isma", "-isma"},
		{".rewrite3gp", "-3gp"},
		{".sampleDesc", "-mpeg4"},
		{".nodrop", "|-nodrop"},
		{".options", ""},
		{0},
	};
	int i, count = 0;
	static const char* argnames[] = {"packed", "mpeg4", "keepsys", "isma", "3gp", 0};
	cmd.Format("\"%s\"", pref.GetValueStr(".path"));
	pref.GenOptions(cmd, opts, ' ');

	if (item->target.video.fps && item->target.video.fpsScale) {
		cmd.AppendFormat(" -fps %f", (float)item->target.video.fps / item->target.video.fpsScale);
	} else if (item->info.video.fps && item->info.video.fpsScale) {
		cmd.AppendFormat(" -fps %f", (float)item->info.video.fps / item->info.video.fpsScale);
	}
	for (i = 0; i < MAX_TRACKS; i++) {
		if (!tracks[i].filename) continue;
		if (tracks[i].type == CT_AUDIO) {
			if (pref.GetValueBool(".sbrx"))
				cmd += " -sbrx";
			else if (pref.GetValueBool(".sbr") || IsSBRAAC(item))
				cmd += " -sbr";
			cmd.AppendFormat(" -add \"%s\"#audio", tracks[i].filename);
		} else if (tracks[i].type == CT_VIDEO) {
			cmd.AppendFormat(" -add \"%s\"#video", tracks[i].filename);
		} else {
			cmd.AppendFormat(" -add \"%s\"", tracks[i].filename);
		}
		if (tracks[i].delay) {
			cmd.AppendFormat(":delay=%d", tracks[i].delay);
		}
		count++;
	}
	cmd.AppendFormat(" \"%s\"", destfile);
	return count == 0 ? 0 : cmd.GetBuffer();
}

char* CMp4box::GetVersionInfo()
{
	CPrefBase prefbase(prefNodePath);
	return GetEncoderVersionInfo(pref.GetValueStr(".path"), "-version", "\n");
}

char* CMp4box::GetOutExt(T_ITEM* item)
{
	CPrefBase prefbase(prefNodePath);
	if (pref.GetValueBool(".rewrite3gp"))
		return "3gp";
	else
		return (mcinst.fAudioEncoding) ? "m4a" : "mp4";
}

int CMp4box::Mux()
{
	memset(&sp ,0, sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	return ShellExec(&sp);
}

int CMp4box::GetProgress()
{ 
	ShellWait(&sp, -1);
	ShellClean(&sp);
	return -1;
}

int CMp4box::ExtractTrack(char* srcfile, CString& destfile, int trackno)
{
	CPrefBase prefbase(prefNodePath);
	CString cmd;
	cmd.Format("\"%s\" \"%s\" -out \"%s\" -raw %d", 
		pref.GetValueStr(".path"),
		srcfile,
		destfile,
		trackno);
	memset(&sp ,0, sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	ShellExec(&sp);
	ShellWait(&sp, -1);
	ShellClean(&sp);
	destfile.AppendFormat("_track%d.*", trackno);
	WIN32_FIND_DATA finddata;
	HANDLE hFind = FindFirstFile(destfile.GetBuffer(), &finddata);
	if (hFind != INVALID_HANDLE_VALUE) {
		destfile.Delete(destfile.GetLength() - 2, 2);
		destfile += strrchr(finddata.cFileName, '.');
		FindClose(hFind);
		return 0;
	} else {
		return -1;
	}
}

bool CPMPMuxer::CheckAudioSupport(int id)
{
	switch (id) {
	case AC_LAMEMP3:
	case AC_NONE:
		return true;
	default:
		return false;
	}
}

bool CPMPMuxer::CheckVideoSupport(int id)
{
	switch (id) {
	case VC_XVID:
	case VC_H264:
		return true;
	default:
		return false;
	}
}

char* CPMPMuxer::UpdateSettings(T_ITEM* item, char* destfile, CONTAINER_TRACKS tracks[])
{
	CPrefBase prefbase(prefNodePath);
	PARAM_OPT opts[] = {
		{".options", ""},
		{0},
	};
	int i;
	bool avc = pref.GetValueInt("overall.video.format") == VC_H264;
	cmd.Format("\"%spmp_muxer%s.exe\"", pref.GetValueStr(".path"), avc ? "_avc" : "");

	for (i = 0; i < MAX_TRACKS; i++) {
		if (!tracks[i].filename) continue;
		switch (tracks[i].type) {
		case CT_AUDIO:
			cmd += " -a ";
			break;
		case CT_VIDEO:
			cmd += " -v ";
			break;
		default:
			continue;
		}
		cmd.AppendFormat("\"%s\"", tracks[i].filename);
	}
	if (avc) {
		cmd.AppendFormat(" -w %d -h %d -r %d -s %d",
			item->target.video.width, item->target.video.height,
			item->target.video.fps ? item->target.video.fps : item->info.video.fps,
			item->target.video.fps ? item->target.video.fpsScale : item->info.video.fpsScale);
	}
	pref.GenOptions(cmd, opts, ' ');
	cmd.AppendFormat(" -o \"%s\"", destfile);
	return cmd.GetBuffer();
}

char* CPMPMuxer::GetVersionInfo()
{
	CPrefBase prefbase(prefNodePath);
	return GetEncoderVersionInfo(pref.GetValueStr(".path"), 0, 0);
}

int CPMPMuxer::Mux()
{
	memset(&sp ,0, sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	return ShellExec(&sp);
}

int CPMPMuxer::GetProgress()
{ 
	ShellWait(&sp, -1);
	ShellClean(&sp);
	return -1;
}

char* CLavf::GetOutExt(T_ITEM* item)
{
	CPrefBase prefbase(prefNodePath);
	static char *ext[] = {"mpg", "asf", "avi", "mp4", "rm", "mov", "flv", "dv"};
	return ext[pref.GetValueInt(".format")];
}

bool CMpegFormat::CheckAudioSupport(int id)
{
	switch (id) {
	case AC_FFMPEG:
	case AC_NONE:
		return true;
	default:
		return false;
	}
}

bool CMpegFormat::CheckVideoSupport(int id)
{
	switch (id) {
	case VC_MPEG1:
	case VC_MPEG2:
		return true;
	default:
		return false;
	}
}
