/*************************************************************************
* MediaCoder - The universal audio/video transcoder
*
* Copyright (c) 2005-06 Stanley Huang <reachme@citiz.net>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. The name of the author may not be used to endorse or promote products
*    derived from this software without specific prior permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************/

#include "config.h"
#include <sys/stat.h>
#include <time.h>
#include "resource.h"
#include "paramlist.h"
#include "processpil.h"
#include "common.h"
#include "transcode.h"
#include "preference.h"
#include "cpuinfo.h"
#include "audioencoders.h"
#include "videoencoders.h"
#include "winamphelper.h"
#include "audiosources.h"
#include "cfilename.h"
#include "mcdialogs.h"
#include "audiofilters.h"
#include "containers.h"
#include "mediatools.h"

#ifdef _HTTPD
#include "httpd.h"
#endif
#include "revision.h"

#define NOW GetTickCount()

CPrefHelper pref("MediaCoderPref");
CAudioEncoder* AudioEncoders[AC_COUNT];
CAudioSource* AudioSources[AUDIO_SRC_COUNT];
CVideoEncoder* VideoEncoders[VE_COUNT];
CContainer* Containers[CF_COUNT];
CONTAINER_TRACKS tracks[MAX_TRACKS];

CAudioSource* audiosrc;
CAudioEncoder* audioenc;
CVideoEncoder* videoenc;
CContainer* container;
#ifdef _HTTPD
CHttpd* httpd = 0;
#endif
MC_STATE inst;
MC_INST_INTERNAL mcinst;
CInstInfo instinfo;

extern "C" INPLUGIN_INFO inplugins;
extern "C" DSPPLUGIN_INFO* dspplugins;
extern "C" int dspcount;
extern T_ITEM itemdef;
extern CVideoEncoder* VideoEncoders[];
extern PARAM_DATA prefParams[];
extern PARAM_DESC prefInfo[];
extern HINSTANCE hDllInst;

int DefaultEventHandler(MC_EVENT event, void* arg);
static OV_ACTION CheckDestExistence(T_ITEM* item, TASK_STATS* stats);
static int GenCommandLines();

int InitCDInterface( int iInterface);
void UninitCDInterface( void );
int LoadParamNames(PARAM_DATA* params, char* filename);

PFN_EVENT_HANDLER pfnEvent = &DefaultEventHandler;

static int timerlevel = 0;
static UINT_PTR timer;

int curEvent;
void* curArg;
char cmd[MAX_PASS_COUNT][1024];
char* prefFile = 0;

static int DefaultEventHandler(MC_EVENT event, void* arg)
{
	curEvent = event;
	curArg = arg;
	SetEvent(mcinst.hEvent);
	return 0;
}

void PopupMsg(int uid)
{
	CString msg;
	msg.LoadString(uid);
	pfnEvent(MC_MSG_POPUP, msg.GetBuffer());
}

void GetAppPath(CString& str)
{
	char buf[MAX_PATH];
	GetModuleFileName(GetModuleHandle(NULL), buf, sizeof(buf));
	str = buf;
	int i = str.ReverseFind('\\');
	if (i != -1) {
		i++;
		str.Delete(i, str.GetLength() - i);
	}
}

FILE_TYPE filetypes[]={
	{"mp3", FT_MP3, "MPEG 1.0 Layer 3", CT_AUDIO},
	{"mp4", FT_MP4, "MPEG4 Audio/Video", CT_VIDEO},
	{"m4a", FT_M4A, "MPEG4 Audio", CT_AUDIO},
	{"3gp", FT_3GP, "3GPP File", CT_VIDEO},
	{"3g2", FT_3GP, "3GPP2 File", CT_VIDEO},
	{"ogg", FT_OGG, "Ogg/Vorbis", CT_AUDIO},
	{"aac", FT_AAC, "MPEG4 AAC", CT_AUDIO},
	{"mp2", FT_MPA, "MPEG 1.0 Layer 2", CT_AUDIO},
	{"mpa", FT_MPA, "MPEG Audio", CT_AUDIO},
	{"wav", FT_WAV, "Waveform", CT_AUDIO},
	{"pcm", FT_WAV, "Raw PCM", CT_AUDIO},
	{"ra", FT_RA, "RealAudio", CT_AUDIO},
	{"ram", FT_RA, "RealAudio", CT_AUDIO},
	{"amr", FT_AMR, "Raw AMR Stream", CT_AUDIO},
	{"mpc", FT_MPC, "MusePack", CT_AUDIO},
	{"wma", FT_WMA, "Windows Media Audio", CT_AUDIO},
	{"ape", FT_APE, "Monkey's Audio", CT_AUDIO},
	{"apl", FT_APE, "Monkey's Audio", CT_AUDIO},
	{"flac", FT_FLAC, "FLAC Audio", CT_AUDIO},
	{"wv", FT_WV, "WavPack Audio", CT_AUDIO},
	{"als", FT_ALS, "MPEG4 Lossless", CT_AUDIO},
	{"cda", FT_CDA, "CD Audio Track", CT_AUDIO},
	{"mpg", FT_MPEG_PS, "MPEG Video", CT_VIDEO},
	{"mpeg", FT_MPEG_PS, "MPEG Video", CT_VIDEO},
	{"mpv", FT_MPEG_PS, "MPEG Video", CT_VIDEO},
	{"mka", FT_MKA, "Matroska Audio", CT_AUDIO},
	{"mkv", FT_MKV, "Matroska Video", CT_VIDEO},
	{"avi", FT_AVI, "Audio Video Interleave", CT_VIDEO},
	{"dat", FT_MPEG_PS, "Video CD Track", CT_VIDEO},
	{"rm", FT_RM, "RealMedia", CT_VIDEO},
	{"rmvb", FT_RM, "RealMedia (VBR)", CT_VIDEO},
	{"mov", FT_MOV, "Quick Time", CT_VIDEO},
	{"wmv", FT_WMV, "Windows Media Video", CT_VIDEO},
	{"asf", FT_WMV, "Windows Media Stream", CT_VIDEO},
	{"vob", FT_MPEG_PS, "DVD/MPEG2", CT_VIDEO},
	{"tp", FT_MPEG_TS, "MPEG2 TS", CT_VIDEO},
	{"ts", FT_MPEG_TS, "MPEG2 TS", CT_VIDEO},
	{"263", FT_263, "Raw H.263 Stream", CT_VIDEO},
	{"264", FT_264, "Raw H.264 Stream", CT_VIDEO},
	{"idx", FT_IDX, "Subtitle", CT_SUBTITLE},
	{"flv", FT_AAC, "Flash Video", CT_VIDEO},
	{"sub", FT_SUB, "Subtitle", CT_SUBTITLE},
	{"cue", FT_CUE, "CUE Sheet", CT_CUE},
	{"m3u", FT_M3U, "CUE Sheet", CT_QUEUE},
	{0},
};

FILE_TYPE ftDVDTrack = {0, FT_MPEG_PS, "DVD Track", CT_VIDEO};
FILE_TYPE ftVCDTrack = {0, FT_MPEG_PS, "Video CD Track", CT_VIDEO};
FILE_TYPE ftCDTrack = {0, FT_CDA, "Audio Track", CT_AUDIO};
FILE_TYPE ftNet = {0, FT_NET, "Net Stream", CT_VIDEO};

FILE_TYPE ftUnknown = {"", FT_UNKNOWN, "", CT_UNKNOWN};

FILE_TYPE* mcGetFileTypeByIndex(int index)
{
	return &filetypes[index];
}

FILE_TYPE* mcIdentifyFileType(char *filename)
{
	char ch;
	int i;
	char *p;
	char *ext;
	if (strstr(filename, "://")) {
		switch (*(LPDWORD)filename) {
		case ':dvd':
			return &ftDVDTrack;
		case ':dvc':
			return &ftVCDTrack;
		case 'addc':
			return &ftCDTrack;
		default:
			return &ftNet;
			
		}
	}
	p = strrchr(filename,'.');
	if (!p) return &ftUnknown;
	ext = strdup(p + 1);
	for (p = ext; (ch = *p); p++) {
		if (ch>='A' && ch<='Z') *p = ch+('a'-'A');
	}
	for (i = 0; filetypes[i].fileext; i++) {
		if (!strcmp(ext, filetypes[i].fileext)) {
			free(ext);
			return &filetypes[i];
		}
	}
	free(ext);
	return &ftUnknown;
}

void PurgeTags(TAGS* tags)
{
	if (tags->title) {
		free(tags->title);
		tags->title = 0;
	}
	if (tags->artist) {
		free(tags->artist);
		tags->artist = 0;
	}
	if (tags->album) {
		free(tags->album);
		tags->album = 0;
	}
	if (tags->genre) {
		free(tags->genre);
        tags->genre = 0;
	}
	if (tags->comment) {
		free(tags->comment);
		tags->comment = 0;
	}
	tags->year = 0;
	tags->valid = 0;
}

int ReadFileTags(T_ITEM* item)
{
	CString cmd;
	SHELL_PARAM sp;
	int bytes;
	if (item->info.filetype->type != CT_AUDIO) return -1;
	PurgeTags(&item->info.tags);
	cmd = "tools\\";
	{
		FILE_TYPE* ft = mcIdentifyFileType(item->filename);
		if (ft && ft->fileext && (!strcmp(ft->fileext, "mp4") || !strcmp(ft->fileext, "m4a"))) {
			cmd += "mp4-tg.exe ";
		} else {
			cmd += "Tag.exe ";
		}
	}
	cmd += '\"';
	cmd += item->filename;
	cmd += '\"';
	memset(&sp,0,sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	sp.flags = SHELL_REDIRECT_STDOUT | SHELL_ALLOC | SHELL_READ_STDOUT_ALL;
	if (ShellExec(&sp)) return -1;
	bytes = ShellRead(&sp, 1000);
	if (bytes > 0) {
		char *p,*v;
		item->info.tags.valid=0;
		strtok(sp.buffer,"\n");
		while ( (p=strtok(0,":")) && (v=strtok(0,"\n")) ) {
			char *t;
			while (*v==' ') v++;
			t=strchr(v,0);
			if ((t--) && *t=='\r') *t=0;
			if (!*v) continue;
			if (!strcmp(p,"Title"))
				item->info.tags.title = strdup(v);
			if (!strcmp(p,"Artist"))
				item->info.tags.artist = strdup(v);
			else if (!strcmp(p,"Album"))
				item->info.tags.album = strdup(v);
			else if (!strcmp(p,"Year"))
				item->info.tags.year = atoi(v);
			else if (!strcmp(p,"Track"))
				item->info.tags.track  =atoi(v);
			//else if (!strcmp(p,"Genre"))
			//	item->info.tags.genre= strdup(v);
			else if (!strcmp(p,"Comment"))
				item->info.tags.comment = strdup(v);
			else
				continue;
			item->info.tags.valid = 1;
		}
	}
	ShellClean(&sp);
	return 0;
}

int WriteFileTags(char* destfile, T_ITEM* item)
{
	SHELL_PARAM sp;
	CString cmd;
	FILE_TYPE* ft;

	if (!item->info.tags.valid) return 0;
	ft = mcIdentifyFileType(destfile);
	cmd = "tools\\";
	if (ft->exttype == FT_M4A || ft->exttype == FT_MP4) {
		cmd.AppendFormat("neroAacTag.exe  \"%s\"", destfile);
		if (item->info.tags.title) {
			cmd.AppendFormat(" -meta:title=\"%s\"", item->info.tags.title);
		}
		if (item->info.tags.artist) {
			cmd.AppendFormat(" -meta:artist=\"%s\"", item->info.tags.artist);
		}
		if (item->info.tags.album) {
			cmd.AppendFormat(" -meta:album=\"%s\"", item->info.tags.album);
		}
		if (item->info.tags.track) {
			cmd.AppendFormat(" -meta:track=\"%d\"", item->info.tags.track);
		}
		if (item->info.tags.year) {
			cmd.AppendFormat(" -meta:year=\"%d\"", item->info.tags.year);
		}
		if (item->info.tags.genre) {
			cmd.AppendFormat(" -meta:genre=\"%s\"", item->info.tags.genre);
		}
		if (item->info.tags.comment) {
			cmd.AppendFormat(" -meta:comment=\"%s\"", item->info.tags.comment);
		}
	} else {
		cmd += "Tag.exe";
		if (item->info.tags.title) {
			cmd.AppendFormat(" --title \"%s\"", item->info.tags.title);
		}
		if (item->info.tags.artist) {
			cmd.AppendFormat(" --artist \"%s\"", item->info.tags.artist);
		}
		if (item->info.tags.album) {
			cmd.AppendFormat(" --album \"%s\"", item->info.tags.album);
		}
		if (item->info.tags.track) {
			cmd.AppendFormat(" --track %d", item->info.tags.track);
		}
		if (item->info.tags.year) {
			cmd.AppendFormat(" --year %d", item->info.tags.year);
		}
		if (item->info.tags.genre) {
			cmd.AppendFormat(" --genre \"%s\"", item->info.tags.genre);
		}
		if (item->info.tags.comment) {
			cmd.AppendFormat(" --comment \"%s\"", item->info.tags.comment);
		}
		cmd.AppendFormat(" \"%s\"", destfile);
	}
	memset(&sp,0,sizeof(sp));
	sp.pchCommandLine = cmd.GetBuffer();
	pfnEvent(MC_MSG_CMD, sp.pchCommandLine);
	if (ShellExec(&sp)) return -1;
	ShellWait(&sp,3000);
	ShellClean(&sp);
	return 0;
}

char* GetEncoderVersionInfo(char* path, char* options, char* delimiter)
{
	static CString verinfo;
	SHELL_PARAM sp;
	memset(&sp,0,sizeof(SHELL_PARAM));
	sp.flags = SHELL_REDIRECT_STDOUT | SHELL_READ_STDOUT_ALL | SHELL_ALLOC;
	verinfo.Format("\"%s\"", path);
	if (options) {
		verinfo += ' ';
		verinfo += options;
	}
	sp.pchCommandLine = verinfo.GetBuffer();
	if (!ShellExec(&sp) && ShellRead(&sp, 3000) > 0) {
		char *p = 0;
		if (delimiter) p = strstr(sp.buffer, delimiter);
		if (!p) p = strstr(sp.buffer, "\r\n\r\n");
		if (!p) p = strstr(sp.buffer, "\n\n");
		if (p) *p = 0;
		verinfo = sp.buffer;
		ShellClean(&sp);
		return verinfo.GetBuffer();
	} else {
		verinfo = (char*)0;
		ShellClean(&sp);
		return 0;
	}
}

void InitRiffHeader(RIFF_HEADER *hdr)
{
	hdr->desc.chunkID='FFIR';
	hdr->desc.chunkSize=0x7ffff024;
	hdr->desc.format='EVAW';
	hdr->fmt.chunkID=' tmf';
	hdr->fmt.chunkSize=16;
	hdr->fmt.wAudioFormat=1;
	hdr->fmt.wChannels=2;
	hdr->fmt.dwSamplesPerSec=44100;
	hdr->fmt.dwAvgBytesPerSec=44100*2*2;
	hdr->fmt.wBlockAlign=2*2;
	hdr->fmt.wBitsPerSample=16;
	hdr->data.chunkID='atad';
	hdr->data.chunkSize=0x7ffff000;
}

void InvalidiateRiffHeader(RIFF_HEADER* hdr, ATTR_AUDIO* audio)
{
	if (audio) {
		hdr->fmt.wBitsPerSample = audio->samplebits;
		hdr->fmt.dwSamplesPerSec = audio->samplerate;
		hdr->fmt.wChannels = audio->channels;
	}
	hdr->fmt.wBlockAlign = hdr->fmt.wBitsPerSample * hdr->fmt.wChannels / 8;
	hdr->fmt.dwAvgBytesPerSec = hdr->fmt.dwSamplesPerSec * hdr->fmt.wBlockAlign;
	hdr->desc.chunkSize = hdr->data.chunkSize + 36;
}

int mcGetEvent(void** parg)
{
	if (WaitForSingleObject(mcinst.hEvent, INFINITE) == WAIT_OBJECT_0) {
		if (parg) *parg = curArg;
		return curEvent;
	}
	return -1;
}

char* mcGetLogText(int maxsize)
{
	return CLogger::GetLog(maxsize);
}

void mcSetLogText(char* logtext)
{
	CLogger::SetLog(logtext);
}

#ifdef _DEBUG
#define HTROOT "..\\mccore\\htdocs"
#else
#define HTROOT "htdocs"
#endif

int mcStartHttpd()
{
#ifdef _HTTPD
	CLogger logger("HTTPd");
	if (!httpd) httpd = new CHttpd;
	inst.httpdPort = pref.GetValueInt("overall.generic.httpdPort");
	CString msg;
	msg.Format("Attempting to start HTTPd on port %d", inst.httpdPort);
	logger.Log(msg);
	inst.httpdPort = httpd->Start(inst.httpdPort, 0, HTROOT, pref.GetValueBool("overall.generic.httpLocalBind"));
	if (!inst.httpdPort) return -1;
	msg.Format("HTTPd started on port %d", inst.httpdPort);
#endif
	return 0;
}

void mcStopHttpd()
{
#ifdef _HTTPD
	CLogger logger("httpd");
	if (httpd) {
		logger.Log("Stopping httpd");
		httpd->Stop();
		delete httpd;
		httpd = 0;
	}
#endif
}

void CALLBACK TimerProc(HWND hWnd, UINT nMsg, UINT nIDEvent, DWORD dwTime)
{
	switch (timerlevel) {
	case 0:
		return;
	case 1:
		(*pfnEvent)(MC_UPDATE_TIME, 0);
		break;
	case 2:
		mcInvalidiateStatsData();
		(*pfnEvent)(MC_UPDATE_STATS, 0);
		break;
	}
}

int mcPostInit(HWND hwndMain)
{
	CString msg;
	CLogger logger;
	
	if (mcGetPrefValueInt("overall.generic.logging"))
		CLogger::Init(mcGetPrefValueStr("overall.generic.logfile"));
	else
		CLogger::Init();
	
	InitCDInterface(2);

	logger.Log(IDS_SCAN_PLUGINS);
	ScanWinampInPlugins(pref.GetValueStr("audiosrc.winamp.path"), &inplugins);
	char* pluginnames = pref.GetValueStr("audiosrc.winamp.disabledPlugins");
	for (INPLUGIN_INFO* cur = inplugins.next; cur; cur = cur->next) {
		for (char *p = pluginnames - 1; p; p = strchr(p, ';')) {
			if (!strncmp(++p, cur->pluginfile, strlen(cur->pluginfile))) {
				SETBIT(cur->flags, FLAG_DISABLED);
				break;
			}
		}
		if (pref.GetValueBool("overall.ui.verbose")) {
			msg.Format("%s - %s (%s)%s\n", cur->pluginfile, cur->desc, cur->exts, 
				ISBITSET(cur->flags, FLAG_DISABLED) ? " [disabled]" : "");
			pfnEvent(MC_MSG_INFO, msg.GetBuffer());
		}
	}
	pluginnames = pref.GetValueStr("audiosrc.winamp.enabledDspPlugins");
	if (ScanDspPlugins(pref.GetValueStr("audiosrc.winamp.path"), &dspplugins)) {
		for (int i = 0; dspplugins[i].hDllPlugin; i++) {
			SETBIT(dspplugins[i].flags, FLAG_DISABLED);
			for (char *p = pluginnames - 1; p; p = strchr(p, ';')) {
				if (!strncmp(++p, dspplugins[i].pluginfile, strlen(dspplugins[i].pluginfile))) {
					CLRBIT(dspplugins[i].flags, FLAG_DISABLED);
					break;
				}
			}
			if (pref.GetValueBool("overall.ui.verbose")) {
				msg.Format("%s - %s%s\n", dspplugins[i].pluginfile, dspplugins[i].hdr->description,
					ISBITSET(dspplugins[i].flags, FLAG_DISABLED) ? " [inactive]" : " [active]");
				pfnEvent(MC_MSG_INFO, msg.GetBuffer());
			}
		}
	}
	LoadDspPlugins(dspplugins, 0);
	mcinst.hwndMain = hwndMain;
	inst.cpucaps.freq = MeasureCPUSpeed(0);

	timer = SetTimer(0, 0, 1000, TimerProc);

	logger.Log("Initialized");
	return 0;
}

void* mcCreateObject(MC_TOOLS type, int index)
{
	switch (type) {
	case TL_TOOL:
		switch (index) {
		case MT_MEDIAINFO:
			return (new CMediaInfo);
		case MT_SHIBATCH:
			return (new CShibatch);
		}
		break;
	}
	return 0;
}

void mcDestroyObject(MC_TOOLS type, int index, void* ptr)
{
	switch (type) {
	case TL_TOOL:
		switch (index) {
		case MT_MEDIAINFO:
			delete (CMediaInfo*)ptr;
			break;
		case MT_SHIBATCH:
			delete (CShibatch*)ptr;
			break;
		}
		break;
	}
}

MC_STATE* mcInit(PFN_EVENT_HANDLER pfnEventCallback)
{
	int i;
	int enumlen = 0;

#ifndef _DEBUG
	{
	char *p, buf[MAX_PATH];
	GetModuleFileName(hDllInst, buf, sizeof(buf));
	p = strrchr(buf, '.');
	strcpy(p + 1, "xml");
	prefFile = strdup(buf);
	}
#else
	prefFile = strdup("..\\mccore\\mccore.xml");
#endif

	memset(&inst, 0, sizeof(MC_STATE));
	memset(&mcinst, 0, sizeof(mcinst));
	mcinst.errmsg = "";
	if (pfnEventCallback) pfnEvent = pfnEventCallback;

	if (pref.Init(prefFile)) {
		mcinst.errmsg = "Preference data cannot be loaded.";
		return 0;
	}

	{
	char buf[MAX_PATH];
	GetTempPath(sizeof(buf), buf);
	pref.SetDefaultValue("overall.task.tempdir", buf, true);
	}

	MeasureCPUSpeed(1);
	GetCpuCaps(&inst.cpucaps);

	// start http deamon if enabled
	if (pref.GetValueBool("overall.generic.httpd")) {
		if (mcStartHttpd()) {
			mcinst.errmsg = "The internal HTTP daemon cannot be started.";
			return 0;
		}
	}

	{
		HDC hScrDC = CreateDC("DISPLAY", NULL, NULL, NULL);
		mcinst.scrw = GetDeviceCaps(hScrDC, HORZRES);
		mcinst.scrh = GetDeviceCaps(hScrDC, VERTRES);
		DeleteDC(hScrDC);
	}

	mcinst.hEvent = CreateEvent(0, FALSE, FALSE, 0);

	inst.state=STATE_IDLE;
	GetAppPath(instinfo.apppath);

	memset(&itemdef, 0, sizeof(T_ITEM));
	itemdef.target.audio = (ATTR_AUDIO*)malloc(sizeof(ATTR_AUDIO));
	itemdef.target.audio->samplerate = 44100;
	itemdef.target.audio->samplebits = 16;
	itemdef.target.audio->channels = 2;

	// create audio encoder objects
	enumlen = 0;
	for (i = 0; i < AC_COUNT; i++) {
		CAudioEncoder* aenc;
		switch (i) {
		case AC_LAMEMP3:	aenc = new CLameMP3; break;
		case AC_3GPPAAC:	aenc = new C3gppAacPlus; break;
		case AC_AACPLUS:	aenc = new CAACPlus; break;
		case AC_HELIX:		aenc = new CHelix; break;
		case AC_FAAC:		aenc = new CFaac; break;
		case AC_VORBIS:		aenc = new CVorbis; break;
		case AC_WAVPACK:	aenc = new CWavPack; break;
		case AC_FLAC:		aenc = new CFlac; break;
		case AC_APE:		aenc = new CApe; break;
		case AC_PCM:		aenc = new CPcm; break;
		case AC_CLI:		aenc = new CGeneric; break;
		case AC_COPY:		aenc = new CAudioCopy; break;
		case AC_NONE:		aenc = new CAudioNone; break;
		case AC_FFMPEG:		aenc = new CFFMpegAudio; break;
		case AC_ALS:		aenc = new CMp4Als; break;
		case AC_OFR:		aenc = new COptimFrog; break;
		case AC_WMA:		aenc = new CWma; break;
		case AC_AMR:		aenc = new CAmr; break;
		case AC_HMP3:		aenc = new CHelixMP3; break;
		case AC_MPC:		aenc = new CMusePack; break;
		case AC_ITUNES:		aenc = new CItunesEnc; break;
		case AC_NEROAAC:	aenc = new CNeroAAC; break;
		case AC_NEROREF:	aenc = new CNeroRefEnc; break;
		case AC_NULL:		aenc = new CNullAudioEncoder; break;
		default:
			mcinst.errmsg = "Error initializing audio codec";
			return 0;
		}
		AudioEncoders[i] = aenc;
	}

	// create audio source objects
	enumlen = 0;
	for (i = 0; i<AS_COUNT; i++) {
		CAudioSource* asrc;
		switch (i) {
		case AS_WINAMP:		asrc = new CWinampInput; break;
		case AS_MPLAYER:	asrc = new CMplayerSource; break;
		case AS_LAME:		asrc = new CLameDecoder; break;
		case AS_WAVE:		asrc = new CWaveFileSource; break;
		case AS_AMR:		asrc = new CAmrDecoder; break;
		default:
			return 0;
		}
		AudioSources[i] = asrc;
	}

	// create video encoder objects
	enumlen = 0;
	for (i = 0; i < VE_COUNT; i ++) {
		CVideoEncoder* venc;
		switch (i) {
		case VE_MENCODER:	venc = new CMencoder; break;
		case VE_FFMPEG:		venc = new Cffmpeg;	break;
		case VE_X264:		venc = new Cx264; break;
		default:
			mcinst.errmsg = "Error initializing video codec";
			return 0;
		}
		VideoEncoders[i] = venc;
	}

	// create container objects
	enumlen = 0;
	for (i = 0; i < CF_COUNT; i ++) {
		CContainer* cf;
		switch (i) {
		case CF_DEFAULT:	cf = new CDefContainer; break;
		case CF_AVI:		cf = new CAVIFormat;	break;
		case CF_MKV:		cf = new CMatroska;		break;
		case CF_MP4:		cf = new CMp4box;		break;
		case CF_MPEG:		cf = new CMpegFormat;	break;
		case CF_LAVF:		cf = new CLavf;			break;
		case CF_PMP:		cf = new CPMPMuxer;		break;
		case CF_RAW:		cf = new CRawContainer;	break;
		default:
			mcinst.errmsg = "Error initializing muxer";
			return 0;
		}
		Containers[i] = cf;
	}

	return &inst;
}

void mcUninit()
{
	CLogger logger("Uninit");
	mcPlayerStop();
	KillTimer(0, timer);
#ifdef _HTTPD
	mcStopHttpd();
#endif
	CloseHandle(mcinst.hEvent);
	CleanupWinampInPlugins(&inplugins);
	UnloadDspPlugins(dspplugins, 1);
	UninitCDInterface();
	free(itemdef.target.audio);
	free(prefFile);
	pref.Destroy();
	CLogger::Uninit();
}

int mcLoadParam(char* cfgfile, int fInit)
{
	if (fInit) {
		pref.Destroy();
		pref.Init(prefFile);
	}
	if (!cfgfile) return 0;
	if (!FileExist(cfgfile)) return -1;
	return pref.Load(cfgfile) ? -1 : 0;
}

int mcSaveParam(char* cfgfile)
{
	if (pref.Save(cfgfile)) {
		return -1;
	} else {
		return 0;
	}
}

void mcApplyParam()
{
	itemdef.target.audioformat = (AUDIO_CODEC)pref.GetValueInt("overall.audio.encoder");
	itemdef.target.videoenc = (VIDEO_ENCODER)pref.GetValueInt("overall.video.encoder");
	itemdef.target.container = (CONTAINER_FORMAT)pref.GetValueInt("overall.video.container");
	itemdef.target.videoformat = (VIDEO_CODEC)pref.GetValueInt("overall.video.format");
	itemdef.target.video.fps = pref.GetValueInt("overall.video.fps");
	itemdef.target.video.fpsScale = pref.GetValueInt("overall.video.fpsScale");
	itemdef.target.video.bitrate = pref.GetValueInt("overall.video.bitrate");
	if (pref.GetValueBool("videofilter.scale.enabled")) {
		itemdef.target.video.width = pref.GetValueInt("videofilter.scale.width");
		itemdef.target.video.height = pref.GetValueInt("videofilter.scale.height");
	}
	VideoEncoders[itemdef.target.videoenc]->UpdateSettings(&itemdef);
	int iaf = itemdef.target.audioformat;
	if (!pref.GetValueBool("overall.ui.verbose")) {
		AudioEncoders[iaf]->flags |= FLAG_QUIET;
	} else {
		AudioEncoders[iaf]->flags &= ~FLAG_QUIET;
	}
	AudioEncoders[iaf]->UpdateSettings(&itemdef);
}

void SetThreadPriorityBySetting(HANDLE hThread)
{
	static int iPriorities[] = {THREAD_PRIORITY_IDLE, THREAD_PRIORITY_LOWEST, THREAD_PRIORITY_BELOW_NORMAL, THREAD_PRIORITY_NORMAL};
	int i = pref.GetValueInt("overall.generic.priority");
	SetThreadPriority(hThread, iPriorities[i < 4 ? i : 3]);
}

void SetProcessPriorityBySetting(HANDLE hProcess)
{
#ifdef CONFIG_WIN9X
	static int iPriorities[] = {IDLE_PRIORITY_CLASS, IDLE_PRIORITY_CLASS, IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS};
#else
	static int iPriorities[] = {IDLE_PRIORITY_CLASS, BELOW_NORMAL_PRIORITY_CLASS, BELOW_NORMAL_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS};
#endif
	int i = pref.GetValueInt("overall.generic.priority");
	SetPriorityClass(hProcess, iPriorities[i < 4 ? i : 3]);
}

void GetVariable(T_ITEM* item, char* varNameString, CString& str)
{
	int i;
	CString varname = varNameString;

	switch (varNameString[0]) {
	case 'D':
		if (varname == "DestFile") {
			bool fFormattedName = false;
			CString value;
			char* destdir = pref.GetValueStr("overall.task.destdir");
			// (DestFile)
			if ((mcinst.fNeedMux && inst.stage != ST_MUXING) || inst.stage == ST_EXTRACTING_PCM) {
				GetVariable(item, "TempFile", value);
			} else if (item->destfile) {
				CFileName fn;
				fn = item->destfile;
				fn.RemoveExtName();
				value += fn.string;
			} else if (*destdir) {
				value += destdir;
				fFormattedName = (pref.GetValueBool("overall.tagging.formatting") && item->info.tags.title);
				if (fFormattedName) {
					QualifyVariables(item, value, pref.GetValueStr("overall.tagging.fileNameFormat"));
				} else {
					if (strstr(item->filename, "://")) {
						GetVariable(item, "SourceFileName", value);
					} else {
						CFileName fn;
						fn = item->filename + mcinst.commonPartLen;
						if (pref.GetValueBool("overall.task.dupDir")) {
							fn.RemoveDriveLetter();
						} else {
							fn.RemovePathName();
						}
						fn.RemoveExtName();
						value += fn.string;
					}
				}
			} else {
				CFileName fn;
				fn = item->filename;
				fn.RemoveExtName();
				value += fn.string;
			}
			
			// add file name postfix
			if (!fFormattedName) {
				if (item->info.track)
					value.AppendFormat(".track%d", item->info.track);
				if (item->info.chapter)
					value.AppendFormat(".chapter%d", item->info.chapter);
			}
			if (mcinst.fileNameSuffix > 0)
				value.AppendFormat(".%d", mcinst.fileNameSuffix);

			//add file extension
			value += '.';
			switch (inst.stage) {
			case ST_EXTRACTING_PCM:
				value += AudioEncoders[AC_PCM]->GetOutExt(item);
				break;
			case ST_PROCESSING_AUDIO:
				if (pref.GetValueBool("overall.audio.allTracks"))
					value.AppendFormat("aid%d.", mcinst.audiochannel);
				else if ((i = pref.GetValueInt("audiofilter.internal.downmix")))
					value.AppendFormat("%c-ch.", i == 1 ? 'l': ( i == 2 ? 'r' : 'm'));
				value += audioenc->GetOutExt(item);
				break;
			case ST_PROCESSING_VIDEO:
				if (container->IsExternalMuxer())
					value += videoenc->GetOutExt(item);
				else
					value += container->GetOutExt(item);
				break;
			case ST_MUXING:
				value += container->GetOutExt(item);
				break;
			}
			if (value == item->filename) {
				i = value.ReverseFind('.');
				value.Insert(i, "_transcoded");
			}
			for (i = 3; i < value.GetLength(); i++) {
				switch (value.GetAt(i)) {
				case ';':
				case ':':
				case '/':
				case '|':
				case '#':
				case '?':
				case '*':
					value.SetAt(i, '_');
					break;
				}
			}
			str += value;
		} else if (varname == "DestDir") {
			str += pref.GetValueStr("overall.task.destdir");
		}
		break;
	case 'S':
		if (varname == "SourceFile") {
			if (inst.stage == ST_PROCESSING_AUDIO) {
				if (mcinst.fUseWaveFile) {
					GetVariable(item, "TempFile", str);
					str.AppendFormat(".%s", AudioEncoders[AC_PCM]->GetOutExt(item));
				} else if (audioenc->flags & FLAG_PIPE) {
					str.AppendFormat("\\\\.\\pipe\\%s%d.wav", OUTPIPENAME, GetCurrentProcessId());
				} else if (!(audioenc->flags & FLAG_STANDALONE)) {
					str += '-';
				} else {
					str += item->filename;
				}
			} else {
				str += item->filename;
				if (pref.GetValueInt("overall.task.mode") == TM_JOIN) {
					for (T_ITEM* itemtmp = item->next; itemtmp; itemtmp = itemtmp->next) {
						str.AppendFormat("\" \"%s", itemtmp->filename);
					}
				}
			}
		} else if (varname == "SourceFileName") {
			char *s = strdup(item->filename), *p;
			p = strstr(s, "://");
			if (p) {
				*p = 0;
				p += 3;
				str += s;
				if (*p) {
					str += '-';
					str += p;
				}
			} else {
				p = strrchr(s, '\\');
				str += p ? p + 1 : s;
			}
		} else if (varname == "SampleRate") {
			str.AppendFormat("%d", item->target.audio->samplerate);
		} else if (varname == "SampleRateKhz") {
			str.AppendFormat("%d", item->target.audio->samplerate / 1000);
		} else if (varname == "SampleBits") {
			str.AppendFormat("%d", item->target.audio->samplebits);
		}
		break;
	case 'T':
		if (varname == "TempFile") {
			str += pref.GetValueStr("overall.task.tempdir");
			GetVariable(item, "SourceFileName", str);
		} else if (varname == "TempDir") {
			str += pref.GetValueStr("overall.task.tempdir");
		} else if (varname == "Title") {
			str += item->info.tags.title ? item->info.tags.title : "Unknown title";
		} else if (varname == "Track") {
			if (item->info.tags.track) 
				str.AppendFormat("%d",item->info.tags.track);
		}
		break;
	case 'V':
		if (varname == "VideoBitrate") {
			int bitrate = pref.GetValueInt("overall.video.bitrate");
			switch (pref.GetValueInt("overall.video.bitrateAdjust")) {
			case 1:
				if (mcinst.bitrate)
					bitrate = (mcinst.bitrate/2/4) + (pref.GetValueInt("overall.video.bitrate")*3/4);
				break;
			case 2:
				if (item->info.duration && item->target.destsize) {
					bitrate = item->target.destsize / (item->info.duration / 1000) * 8;
				}
				break;
			case 3:
				if (item->info.duration && item->info.filesize && item->target.ratio) {
					bitrate = (int)((float)(item->info.filesize / 1024 / (item->info.duration / 1000) * 8) / item->target.ratio);
				}
				break;
			}
			str.AppendFormat("%d", bitrate);
		}
		break;
	case 'C':
		if (varname == "ChannelNum") {
			str.AppendFormat("%d", item->target.audio->channels);
		} else if (varname == "Comment") {
			str += item->info.tags.comment;
		}
		break;
	case 'A':
		if (varname == "AudioFiles") {
			for (i = 0; i<MAX_TRACKS; i++) {
				if (tracks[i].filename && tracks[i].type == CT_AUDIO) {
					str += tracks[i].filename;
					str += ' ';
				}
			}
			str.Delete(str.GetLength() - 1);
		} else if (varname == "AppPath") {
			str += instinfo.apppath;
		} else if (varname == "Artist") {
			str += item->info.tags.artist ? item->info.tags.artist : "Unknown artist";
		} else if (varname == "Album") {
			str += item->info.tags.album;
		} else if (varname == "AudioBytes") {
			int bytesPerSec = item->target.audio->samplerate * item->target.audio->channels * item->target.audio->samplebits / 8;
			str.AppendFormat("%d", MS2BYTE(item->info.duration, bytesPerSec));
		}
		break;
	case 'G':
		if (varname == "Genre") {
			str += item->info.tags.genre;
		}
		break;
	case 'Y':
		if (varname == "Year") {
			if (item->info.tags.year)
				str.AppendFormat("%d",item->info.tags.year);
		}
		break;
	case 'N':
		if (varname.Left(3) == "Now") {
			CString subname = varname.Mid(3);
			time_t now = time(0);
			struct tm *btm = localtime(&now);
			int n = -1;
			if (subname == "Year") {
				str.AppendFormat("%04d", 1900+btm->tm_year);
			} else if (subname == "Month") {
				n = btm->tm_mon;
			} else if (subname == "Day") {
				n = btm->tm_mday;
			} else if (subname == "WeekDay") {
				n = btm->tm_wday;
			} else if (subname == "Hour") {
				n = btm->tm_hour;
			} else if (subname == "Min") {
				n = btm->tm_min;
			} else if (subname == "Sec") {
				n = btm->tm_sec;
			} else {
				n = -1;
			}
			if (n >= 0) {
				str.AppendFormat("%02d", n);
			}
		}
		break;
	}
}

int EncodeAudioData(char* buf, int len, T_ITEM* item)
{
	int bytes = len;
	MC_AUDIO_FILTER *af;
	// filter audio data if needed
	for (af = mcinst.audiofilter; af; af = af->next) {
		bytes = (*mcinst.audiofilter->pfnAudioDsp)((short*)buf, len >> 1) << 1;
	}

	// filter audio with DSP plugins;
	if (dspcount > 0) {
		int srate = item->target.audio->samplerate;
		int channels = item->target.audio->channels;
		int bits = item->target.audio->samplebits;
		int samplenum = len * 8 / channels / bits;
		for (int i = 0; dspplugins[i].hDllPlugin; i++) {
			if (dspplugins[i].ModifySamples)
				dspplugins[i].ModifySamples(dspplugins[i].mod, (short*)buf, samplenum, bits, channels, srate);
		}
	}

	// send to encoder
	bytes = audioenc->Encode(buf, bytes);
	return bytes;
}


static int WaitAudioEncoder(int timeout)
{
	if (audioenc->flags & FLAG_STANDALONE) {
		while (inst.state==STATE_ENCODING && (inst.astats.percent = (float)audioenc->GetProgress()) != -1);
	} else {
		int aliveCountPrev = -1;
		for (;;) {
			if (inst.state == STATE_ENCODING) {
#ifdef _DEBUG
				char buf[16];
				sprintf(buf, "%d\r", mcinst.aliveCounter);
				pfnEvent(MC_MSG_INFO, buf);
#endif
				if (WaitForSingleObject(mcinst.hPumpThread, timeout) != WAIT_TIMEOUT) break;
				if (mcinst.aliveCounter == aliveCountPrev) {
					// the pump is blocked
					TerminateThread(mcinst.hPumpThread, 0);
					return 0;
				}
				aliveCountPrev = mcinst.aliveCounter;
			} else {
				if (WaitForSingleObject(mcinst.hPumpThread, timeout) == WAIT_TIMEOUT)
					TerminateThread(mcinst.hPumpThread, 0);
				return -1;
			}
		}
	}
	return 0;
}

static void AudioDataPump(T_ITEM* item)
{
	int bytesOut = 0;
	int bytesIn;
	int bytesPerSec = item->target.audio->samplerate * item->target.audio->channels * item->target.audio->samplebits / 8;

	if (mcinst.audioDelay > 0) {
		int padbytes = (int)MS2BYTE(mcinst.audioDelay, bytesPerSec);
		memset(mcinst.audiodata, 0, mcinst.audioFrameSize);
		for (int i = 0; 
			i < padbytes - mcinst.audioFrameSize; 
			i += EncodeAudioData(mcinst.audiodata, mcinst.audioFrameSize, item));
		if (i < padbytes) EncodeAudioData(mcinst.audiodata, padbytes - i, item);
	} else if (mcinst.audioDelay < 0){
		int padbytes = (int)MS2BYTE(-mcinst.audioDelay, bytesPerSec);
		bytesOut = 0;
		for (int i = 0; i < padbytes - mcinst.audioFrameSize; i += bytesOut) {
			bytesOut = audiosrc->read(mcinst.audiodata, mcinst.audioFrameSize);
			if (bytesOut < 0) break;
		}
		if (bytesOut < 0)
			inst.state = STATE_ERROR;
		else
			if (i < padbytes) audiosrc->read(mcinst.audiodata, padbytes - i);
		if (audioenc->framesize) mcinst.audioFrameSize = audioenc->framesize * 2;
	}
	if (inst.state == STATE_ERROR) return;
	while (inst.state==STATE_ENCODING && 
			audiosrc->bytescount < audiosrc->totalbytes &&
			(bytesIn = audiosrc->read(mcinst.audiodata, mcinst.audioFrameSize)) >= 0) {
		bytesOut = EncodeAudioData(mcinst.audiodata, bytesIn, item);
		if (bytesOut<=0) break;;
		mcinst.aliveCounter++;
	}
	//logger.Log(IDS_UNLOADED, audiosrc->GetName());
}

static int StartTranscodeAudio(T_ITEM* item, HANDLE hInput)
{
	int ret=0;
	CLogger logger;
	CString msg;
	T_ITEM* itemint = 0;

	inst.stage = ST_PROCESSING_AUDIO;
	item->target.audio = itemdef.target.audio;

	mcinst.audioFrameSize = AUDIO_BLOCK_SIZE;
	mcinst.audioDelay = item->info.audiodelay ? item->info.audiodelay : pref.GetValueInt("overall.audio.delay");

	memset(&inst.astats, 0, sizeof(AUDIO_STATS));
	instinfo.destfile = "";
	GetVariable(item, "DestFile", instinfo.destfile);
	instinfo.audiofile = instinfo.destfile;
	pfnEvent(MC_UPDATE_STATE, (void*)inst.state);

	audiosrc = AudioSources[pref.GetValueInt("overall.audio.source")];
	if (inst.stage == ST_EXTRACTING_PCM) {
		audiosrc = AudioSources[AS_WAVE];
	} else {
		if (item->info.filetype->type == CT_VIDEO) {
			audiosrc = AudioSources[AS_MPLAYER];
		} else {
			switch (item->info.filetype->exttype) {
			case FT_CDA:
			case FT_MPC:
			case FT_WV:
			case FT_APE:
			case FT_FLAC:
			case FT_ALS:
				audiosrc = AudioSources[AS_WINAMP];
				break;
			}
		}
	}

	// check if the encoder can decode source audio file
	mcinst.fAudioDecodable = false;
	if (audioenc->GetInExts()) {
		char* exts = strdup(audioenc->GetInExts());
		char* p = strtok(exts, ";");
		for ( ; p; p = strtok(0, ";")) {
			if (item->info.filetype->fileext && !strcmp(item->info.filetype->fileext, p)) {
				mcinst.fAudioDecodable = true;
				break;
			}
		}
		free(exts);
	}

	if (pref.GetValueInt("overall.audio.source") == AS_AMR) {
		CString source;
		GetVariable(item, "TempDir", source);
		source += "mp4boxout";
		Containers[CF_MP4]->ExtractTrack(item->filename, source, 2);
		itemint = mcMakeItem(source.GetBuffer());
	} else if (mcinst.fUseWaveFile) {
		CString source;
		pfnEvent(MC_UPDATE_STATE, (void*)inst.state);
		pfnEvent(MC_UPDATE_TITLE, "extracting waveform");
		logger.Log(IDS_EXTRACT_WAVE, item->filename);
		GetVariable(item, "SourceFile", source);
		timerlevel = 1;
		if (audiosrc->extract(item, source.GetBuffer(), mcinst.startpos)) return -1;
		if (pref.GetValueBool("audiofilter.shibatch.enabled")) {
			// run shibatch tool to process the wave file
			int ret;
			CMediaTool *mt = (CMediaTool*)mcCreateObject(TL_TOOL, MT_SHIBATCH);
			if (mt) {
				T_ITEM item;
				pfnEvent(MC_UPDATE_TITLE, "processing waveform");
				item.filename = source.GetBuffer();
				ret = mt->Run(0, &item);
				pfnEvent(MC_MSG_INFO, mt->result.GetBuffer());
				mcDestroyObject(TL_TOOL, MT_SHIBATCH, mt);
			}
		}
		audiosrc = AudioSources[AS_WAVE];
		itemint = mcMakeItem(source.GetBuffer());
	}

	// opening audio source
	logger.Log(IDS_LOAD_AUDIO_SOURCE, audiosrc->GetName());
	audiosrc->samplerate = 0;
	{
		int low = audioenc->GetLowestSampleRate();
		int high = audioenc->GetHighestSampleRate();
		if (low == high && low != 0) {
			audiosrc->samplerate = low;
		}
	}
	
	audiosrc->duration = mcinst.duration;
	audiosrc->hInput = hInput;
	ret = audiosrc->open(itemint ? itemint : item, mcinst.startpos);

	if (itemint)
		mcFreeItem(itemint);
	if (ret) return -1;

	memcpy(item->target.audio, &item->info.audio, sizeof(ATTR_AUDIO));
	if (audiosrc->samplerate) item->target.audio->samplerate = audiosrc->samplerate;
	if (audiosrc->samplebits) item->target.audio->samplebits = audiosrc->samplebits;
	if (audiosrc->channels) item->target.audio->channels = audiosrc->channels;

	msg.Format("\nAudio source: %s\nSample Rate: %dHz Channels: %d BPS: %d\n",
		audiosrc->GetName(), audiosrc->samplerate, audiosrc->channels, audiosrc->samplebits);
	pfnEvent(MC_MSG_INFO, msg.GetBuffer());

	inst.astats.length = mcinst.duration > 0 ? mcinst.duration : (item->info.duration - mcinst.startpos);

	MC_AUDIO_FILTER af;
	mcinst.audiofilter = 0;
	if (item->info.audio.channels != 1) {
		int i = pref.GetValueBool("audiofilter.internal.downmix");
		if (i) {
			memset(&af, 0, sizeof(af));
			if (pref.GetValueBool("overall.audio.allTracks")) {
				af.pfnAudioDsp = dspLeftChannelMono;
				af.channels = 1;
			} else {
				switch (i) {
				case 1: af.pfnAudioDsp = dspLeftChannelMono; break;
				case 2: af.pfnAudioDsp = dspRightChannelMono; break;
				default: af.pfnAudioDsp = dspDownmix; break;
				}
				af.channels = 1;
			}
			mcinst.audiofilter = &af;
		} else if (audioenc->GetMaxChannel() == 1) {
			memset(&af, 0, sizeof(af));
			af.pfnAudioDsp = dspDownmix;
			af.channels = 1;
			mcinst.audiofilter = &af;
		}
	}

	int srate = pref.GetValueInt("overall.decoding.forceSampleRate");
	if (srate) item->target.audio->samplerate = srate;
	if (audioenc->framesize)
		mcinst.audioFrameSize = audioenc->framesize;

	// adjust audio parameters by applied audio filters
	if (mcinst.audiofilter) {
		MC_AUDIO_FILTER* af = mcinst.audiofilter;
		if (af->channels)
			item->target.audio->channels = af->channels;
		if (af->samplerate)
			item->target.audio->samplerate = af->samplerate;
		if (af->bps)
			item->target.audio->samplebits = af->bps;
		if (item->info.audio.channels > item->target.audio->channels) {
			mcinst.audioFrameSize = mcinst.audioFrameSize * item->info.audio.channels / item->target.audio->channels;
			audiosrc->framesize = mcinst.audioFrameSize;
		}
	}
	if (item->target.audio->samplerate < audioenc->GetLowestSampleRate() ||
		item->target.audio->samplerate> audioenc->GetHighestSampleRate()) {
		logger.Log(IDS_INVALID_SAMPLE_RATE, audioenc->GetFullName());
		return -1;
	}

	msg.Format("\nAudio encoder: %s\nSample Rate: %dHz Channels: %d BPS: %d\n\n",
		audioenc->GetFullName(), item->target.audio->samplerate, item->target.audio->channels, item->target.audio->samplebits);
	pfnEvent(MC_MSG_INFO, msg.GetBuffer());

	msg = "";
	audioenc->UpdateSettings(item);
	QualifyVariables(item, msg, audioenc->cl.GetBuffer());
	logger.Log(IDS_LOAD_AUDIO_ENCODER, audioenc->GetFullName());
	if (audioenc->Open(msg, instinfo.destfile, item)) {
		logger.Log(IDS_ERROR_LOADING, audioenc->GetFullName());
		audiosrc->close();
		return -1;
	}
	mcContainerAddFile(CT_AUDIO, instinfo.destfile.GetBuffer(), DUMMY_FLAG);

	if (!(audioenc->flags & FLAG_RAW_PCM)) {
		RIFF_HEADER hdr;
		InitRiffHeader(&hdr);
		InvalidiateRiffHeader(&hdr, item->target.audio);
        audioenc->Encode((char*)&hdr,sizeof(RIFF_HEADER));
	}

	if ((audioenc->flags & FLAG_STANDALONE)) {
		mcinst.hPumpThread = 0;
	} else {
		DWORD dwid;
		mcinst.filterbuf = (char*)malloc(mcinst.audioFrameSize);
		mcinst.audiodata = (char*)malloc(mcinst.audioFrameSize);
		mcinst.aliveCounter = 0;
		mcinst.hPumpThread = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)AudioDataPump, (void*)item, 0, &dwid);
		SetThreadPriority(mcinst.hPumpThread, THREAD_PRIORITY_LOWEST);
	}
	return ret;
}

void FinishTranscodeAudio(T_ITEM* item)
{
	CLogger logger;
	timerlevel = 0;
	if (mcinst.hPumpThread) {
		CloseHandle(mcinst.hPumpThread);
		mcinst.hPumpThread = 0;
	}
	if (mcinst.filterbuf) free(mcinst.filterbuf);
	if (mcinst.audiodata) free(mcinst.audiodata);

	audiosrc->close();
	logger.Log(IDS_UNLOADED, audiosrc->GetName());
	audioenc->Close();
	logger.Log(IDS_UNLOADED, audioenc->GetFullName());
	if ((audioenc->flags & FLAG_TO_TAG) && pref.GetValueBool("overall.tagging.tag") && !mcinst.fNeedMux) {
		WriteFileTags(instinfo.destfile.GetBuffer(), item);
	}
	if (mcinst.fUseWaveFile && !pref.GetValueBool("overall.task.keepTemp")) {
		CString tempfile;
		GetVariable(item, "TempFile", tempfile);
		tempfile.AppendFormat(".%s", AudioEncoders[AC_PCM]->GetOutExt(item));
		RemoveFile(tempfile.GetBuffer(), 1000);
	}
	mcinst.fAudioDecodable = false;
	mcinst.fUseWaveFile = false;
	inst.astats.percent = 100;
	pfnEvent(MC_UPDATE_PROGRESS, (void*)&inst.astats);

}

char* mcContainerGetFile(TRACK_TYPE type, int num)
{
	int n = 0;
	for (int i = 0; i < MAX_TRACKS; i++) {
		if (tracks[i].type == type) {
			if (n == num) {
				return tracks[i].filename;
			}
			n++;
		}
	}
	return 0;
}

int mcContainerAddFile(TRACK_TYPE type, char* filename, CF_FLAGS flags)
{
	int i;
	if (!filename) return -1;
	for (i=0;i<MAX_TRACKS;i++) {
		if (!tracks[i].filename) {
			tracks[i].filename=strdup(filename);
			if (type==CT_UNKNOWN) {
				FILE_TYPE* ft = mcIdentifyFileType(filename);
				if (ft) {
					type = ft->type;
				}
			}
			tracks[i].type=type;
			tracks[i].flags=flags;
			return 0;
		}
	}
	return -1;
}

int mcContainerRemoveFile(char* filename)
{
	int i;
	if (!filename) return -1;
	for (i=0;i<MAX_TRACKS;i++) {
		if (tracks[i].filename && !strcmp(tracks[i].filename,filename)) {
			free(tracks[i].filename);
			tracks[i].filename=0;
			tracks[i].type=CT_UNKNOWN;
			return 0;
		}
	}
	return -1;
}

int mcContainerPurge(T_ITEM* item, int fDelete)
{
	int i;
	CString tmpfile;
	GetVariable(item, "TempFile", tmpfile);
	for (i=0; i<MAX_TRACKS; i++) {
		if (tracks[i].filename) {
			if (fDelete && !strncmp(tracks[i].filename, tmpfile.GetBuffer(), tmpfile.GetLength())) DeleteFile(tracks[i].filename);
			free(tracks[i].filename);
			memset(&tracks[i], 0, sizeof(CONTAINER_TRACKS));
			tracks[i].type=CT_UNKNOWN;
		}
	}
	return 0;
}

int mcContainerMuxFiles(T_ITEM* item, char* destfile)
{
	if (item->target.container == CF_DEFAULT) return -1;
	char *p = Containers[item->target.container]->UpdateSettings(item, destfile, tracks);
	if (!p) return 0;
	int percent = 0;
	pfnEvent(MC_MSG_CMD, p);
	if (Containers[item->target.container]->Mux()) {
		CLogger logger;
		logger.Log(IDS_ERROR_MUXER);
		return -1;
	}
	pfnEvent(MC_UPDATE_PROGRESS, 0);
	while ((percent = Containers[item->target.container]->GetProgress()) >= 0) {
		if (percent > 0) {
			pfnEvent(MC_UPDATE_PROGRESS, (void*)percent);
		}
	}
	if (pref.GetValueBool("overall.tagging.tag") && item->info.tags.valid) {
		return WriteFileTags(instinfo.destfile.GetBuffer(), item);
	}
	return 0;
}

void QualifyVariables(T_ITEM* item, CString& cmd, char *in)
{
	char *q;

	for (q = in; *q; ) {
		if (*q=='$') {
			if (*(q+1)=='(') {
				char *s = strchr(q+2,')');
				if (s) {
					*s=0;
					GetVariable(item, q+2, cmd);
					*s=')';
					//bool fQuote=(*q!='\"' && *(s+1)!='\"' && varbuf[0]!='\"' && strchr(varbuf,' '));
					//p += sprintf(p,fQuote?"\"%s\"":"%s",varbuf);
					q = s + 1;
					continue;
				}
			}
		}
		cmd += *(q++);
	}
}

static int MakeDir(char* dirname)
{
	char *p=dirname;
	DWORD attr;
	if (dirname[1]==':') p+=2;
	for(;;) {
		p=strchr(p+1,'\\');
		if (!p) {
			attr=GetFileAttributes(dirname);
			if (attr == INVALID_FILE_ATTRIBUTES) {
				return CreateDirectory(dirname,NULL) ? 0 : -1;
			} else {
				return (attr & FILE_ATTRIBUTE_DIRECTORY) ? 0 : -1;
			}
		}
		*p=0;
		attr=GetFileAttributes(dirname);
		if (attr == INVALID_FILE_ATTRIBUTES) {
			if (!CreateDirectory(dirname,NULL)) return -1;
		} else if (!(attr & FILE_ATTRIBUTE_DIRECTORY)) {
			return -1;
		}
		*p='\\';
	}
}

OV_ACTION CheckDestExistence(T_ITEM* item, TASK_STATS* stats)
{
	if (!item || !item->destfile) return OA_SKIP;
	char* filename = item->destfile;
	OV_ACTION choice = (OV_ACTION)pref.GetValueInt("overall.task.overwrite");
	if (FileExist(filename)) {
		CString msg;
		if (choice == OA_SKIP) {
			return OA_SKIP;
		}
		if (choice == OA_ASK) {
			if (mcinst.ovaction == OA_ASK) {
				msg.Format(IDS_OVERWRITE_CONFIRM, filename);
				switch (pfnEvent(MC_MSG_YESNOCANCEL, msg.GetBuffer())) {
				case IDYES:
					choice = OA_OVERWRITE;
					break;
				case IDNO:
					choice = OA_BACKUP;
					break;
				default:
					choice = OA_SKIP;
					break;
				}
				if (item->next) {
					msg.Format(IDS_OVERWRITE_KEEP, filename);
					if (pfnEvent(MC_MSG_YESNO, msg.GetBuffer()) == IDYES) {
						mcinst.ovaction = choice;
					}
				}
			} else {
				choice = mcinst.ovaction;
			}
		}
		switch (choice) {
		case OA_BACKUP: {
			msg = filename;
			int i = msg.ReverseFind('.');
			if (i > 0)
				msg.Insert(i, ".bak");
			else
				msg += ".bak";
			MoveFile(filename, msg.GetBuffer());
			} break;
		case OA_OVERWRITE:
			if (RemoveFile(filename, 1000)) {
				msg.Format(IDS_ERROR_DELETE, filename);
				pfnEvent(MC_MSG_ALERT, msg.GetBuffer());
				choice = OA_SKIP;
			}
			break;
		}
	} else {
		choice = OA_NOTEXIST;
	}
	return choice;
}

static int TranscodeVideo(T_ITEM* item)
{
	CString cmd;
	CString msg;
	int ret = 0;
	HANDLE hPipeRead, hPipeWrite = 0;
	CLogger logger;
	mcinst.bitrate = 0;
	inst.stage=ST_PROCESSING_VIDEO;

	if (item->target.video.fps && pref.GetValueBool("overall.video.nofpsup")) {
		if (GetFPS(item->info.video.fps, item->info.video.fpsScale) < GetFPS(item->target.video.fps, item->target.video.fpsScale)) {
			item->target.video.fps = 0;
		}
	}

	videoenc->UpdateSettings(item);
	inst.totalpass = videoenc->GetPassCount();
	pfnEvent(MC_UPDATE_STATE, (void*)inst.state);
	//GenCommandLines();
	inst.totalpass = videoenc->GetPassCount();
	for (inst.currentpass = 1; inst.currentpass <= inst.totalpass && inst.state==STATE_ENCODING; inst.currentpass++) {
		timerlevel = 1;
		if (videoenc->open()) {
			logger.Log(IDS_ENCODER_FAIL, videoenc->GetName());
			break;
		}

		inst.stats.dwClipStartTime = NOW;
		logger.Log(IDS_LOADED, videoenc->GetName());
		memset(&inst.vstats, 0, sizeof(VIDEO_STATS));
		cmd.Format("pass %d of %d", inst.currentpass, inst.totalpass);
		pfnEvent(MC_UPDATE_TITLE, cmd.GetBuffer());

		cmd = "";
		QualifyVariables(item, cmd, videoenc->GetCommand(inst.currentpass - 1));

		if (inst.currentpass == inst.totalpass && !mcinst.fSoleAudio && mcinst.fNeedMux) {
			if (!CreatePipe(&hPipeRead, &hPipeWrite, 0, AUDIO_BLOCK_SIZE)) return -1;
			SetHandleInheritance(&hPipeWrite, TRUE);
			cmd.Replace(" -nosound", "");
			cmd.AppendFormat(" -oac pcm -pcmopts pipe=%d:pipeonly", hPipeWrite);
		}
		//start external encoder
		if (videoenc->Encode(cmd)) {
			videoenc->close();
			ret = -1;
			logger.Log(IDS_UNLOADED, videoenc->GetName());
			break;
		}
		if (hPipeWrite) {
			int ret = StartTranscodeAudio(item, hPipeRead);
			inst.stage = ST_PROCESSING_VIDEO;
			instinfo.destfile = "";
			GetVariable(item, "DestFile",instinfo.destfile);
			CloseHandle(hPipeWrite);
			if (ret) {
				hPipeWrite = 0;
			}
		}
		//wait for encoder to finish
		timerlevel = 2;
		switch (pref.GetValueInt("overall.task.mode")) {
		case TM_NORMAL:
		case TM_SPLIT:
			// normal mode
			while ( inst.state == STATE_ENCODING && videoenc->GetProgress() >= 0);
			break;
		case TM_JOIN:
			// joiner mode
			int prev = 0, cur;
			T_ITEM* workitem = item;
			while (inst.state == STATE_ENCODING) {
				cur = videoenc->GetProgress();
				if (cur < 0) break;
				if (cur < prev && workitem) {
					pfnEvent(MC_JOB_FINISH, workitem);
					workitem = workitem->next;
					if (workitem) {
						pfnEvent(MC_JOB_START, workitem);
						inst.stats.decodedMB += (unsigned int)(workitem->info.filesize >> 20);
					}
				}
				prev = cur;
			}
			break;
		}
		mcinst.bitrate = inst.vstats.vbitrate;
		if (hPipeWrite) {
			timerlevel = 0;
			WaitAudioEncoder(1000);
			FinishTranscodeAudio(item);
		}
		videoenc->close();
		logger.Log(IDS_UNLOADED, videoenc->GetName());
	}
	if (ret) return ret;
	instinfo.destfile = "";
	GetVariable(item, "DestFile",instinfo.destfile);
	mcContainerAddFile(CT_VIDEO, instinfo.destfile.GetBuffer(), DUMMY_FLAG);
	return 0;
}

int mcStopTranscode()
{
	if (inst.state == STATE_IDLE) return -1;
	inst.state = STATE_STOPPING;
	return WaitForSingleObject(mcinst.hEvent, 3000) == WAIT_OBJECT_0 ? 0 : -1;
}

int mcTranscodeItem(T_ITEM* item)
{
	bool fAllTracks = pref.GetValueBool("overall.audio.allTracks");
	bool fVerbose = pref.GetValueBool("overall.ui.verbose");
	CString str;
	CLogger logger;

	// check accessibility of working folder
	{
		DWORD attr;
		str = pref.GetValueStr("overall.task.tempdir");
		if (!str.IsEmpty()) {
			if (str.Right(1) == '\\') str.Delete(str.GetLength() - 1);
			attr = GetFileAttributes(str.GetBuffer());
			if (attr == INVALID_FILE_ATTRIBUTES || !(attr & FILE_ATTRIBUTE_DIRECTORY)) {
				str.Format(IDS_INVALID_TEMP_DIR);
				pfnEvent(MC_MSG_ALERT, str.GetBuffer());
				return -1;
			}
		}
	}

	inst.state = STATE_ENCODING;
	DWORD tm = NOW;
	inst.stats.dwTaskStartTime = tm;
	memset(&inst.vstats, 0, sizeof(VIDEO_STATS));
	instinfo.audiofile = "";

	mcGetMediaInfo(item);
	mcApplyParam();
	if (!(item->flags & IF_PRIVATE_FORMAT)) {
		item->target.audioformat = itemdef.target.audioformat;
		item->target.videoformat = itemdef.target.videoformat;
		item->target.videoenc = itemdef.target.videoenc;
		item->target.container = itemdef.target.container;
		item->target.video.fps = itemdef.target.video.fps;
		item->target.video.fpsScale = itemdef.target.video.fpsScale;
		item->target.video.bitrate = itemdef.target.video.bitrate;
	}
	if (!item->target.video.width) item->target.video.width = itemdef.target.video.width;
	if (!item->target.video.height) item->target.video.height = itemdef.target.video.height;

	audioenc = AudioEncoders[item->target.audioformat];
	videoenc = VideoEncoders[item->target.videoenc];
	container = Containers[item->target.container];
	mcinst.fAudioEncoding = (item->target.videoformat == VC_NONE || item->info.filetype->type == CT_AUDIO);

	if (mcinst.startTimeOffset == 0) {
		int splitParam = pref.GetValueInt("overall.task.splitParam");
		logger.Log(IDS_START_JOB, item->filename);
		inst.stats.fileprocessed++;
		pfnEvent(MC_JOB_START, item);
		if ((item->flags & IF_DONE) && mcinst.taskMode != TM_JOIN) {
			inst.state = STATE_SKIPPING;
			goto skip;
		}
		if (item->target.splitDuration > 0) {
            mcinst.splitTime = item->target.splitDuration;
		} else if (splitParam && mcinst.taskMode == TM_SPLIT) {
			if (pref.GetValueInt("overall.task.splitMode") == 0) {
				mcinst.splitTime = splitParam;
			} else {
				int vbitrate = videoenc->GetTargetBitRate(item);
				int abitrate = audioenc->GetTargetBitRate(item);
				if (abitrate < 0 || vbitrate < 0) {
					inst.state = STATE_ERROR;
					PopupMsg(IDS_SPLIT_ERROR);
					goto skip;
				}
				mcinst.splitTime = splitParam * 8000 / (vbitrate + abitrate);
			}
		} else {
			mcinst.splitTime = 0;
		}
        if (mcinst.splitTime > item->info.duration) mcinst.splitTime = 0;
		mcinst.fileNameSuffix = mcinst.splitTime ? 1 : 0;
	}
	mcinst.startpos = item->info.begin ? item->info.begin : pref.GetValueInt("overall.decoding.startTime") * 1000;
	if (mcinst.splitTime) {
		mcinst.startpos += mcinst.startTimeOffset;
		mcinst.duration = mcinst.splitTime;
	} else {
		mcinst.duration = item->info.end ? item->info.end - mcinst.startpos : pref.GetValueInt("overall.decoding.duration") * 1000;
	}

	/*
	if (fVerbose && 
		((mcinst.fAudioEncoding && !item->info.duration) ||
		!mcinst.fAudioEncoding && (!item->info.video.codec || !item->info.video.width || !item->info.video.height))) {
		int ret;
		str.Format(IDS_INVALID_FILE, item->filename);
		ret = pfnEvent(MC_MSG_YESNO, str.GetBuffer());
		if (ret == IDNO) {
			inst.state = STATE_SKIPPING;
			goto skip;
		}
	}
	*/

	mcinst.fNeedMux = false;
	if (item->target.container != CF_DEFAULT) {
		if (!videoenc->CheckContainerSupport(item->target.container)) {
			if (audioenc->flags & FLAG_NO_ENCODER) {
				logger.Log(IDS_AUDIO_STREAM_UNSUPPORTED);
				inst.state = STATE_ERROR;
			} else {
				mcinst.fNeedMux = true;
			}
		}
	}

	mcinst.fUseWaveFile = ((audioenc->flags & FLAG_STANDALONE) && !mcinst.fAudioDecodable) || pref.GetValueBool("overall.audio.useWaveFile") || pref.GetValueBool("audiofilter.shibatch.enabled");

	if (mcinst.fAudioEncoding || fAllTracks || mcinst.fUseWaveFile) {
		mcinst.fSoleAudio = true;
	} else if (!mcinst.fNeedMux) {
		mcinst.fSoleAudio = false;
	} else if (pref.GetValueInt("overall.audio.source") != AS_MPLAYER || !strcmp(videoenc->GetOutExt(item), "avi")) {
		mcinst.fSoleAudio = true;
	} else {
		mcinst.fSoleAudio = pref.GetValueBool("overall.task.audioSolely");
	}

	if (!container->CheckAudioSupport(item->target.audioformat) || !container->CheckVideoSupport(item->target.videoformat)) {
		if (fVerbose) PopupMsg(IDS_ERROR_MUX);
		inst.state = STATE_ERROR;
	}
	if (inst.state!=STATE_ENCODING) goto skip;

	// check for destination file existance
	inst.stage = ST_MUXING;
	if (!item->destfile) {
		instinfo.destfile = "";
		GetVariable(item, "DestFile", instinfo.destfile);
		mcSetItemStr(&item->destfile, instinfo.destfile.GetBuffer());
	} else {
		instinfo.destfile = item->destfile;
	}
	switch (CheckDestExistence(item, &inst.stats)) {
	case OA_SKIP:
		inst.state = STATE_SKIPPING;
		goto skip;
	case OA_NOTEXIST: {
		int slash = instinfo.destfile.ReverseFind('\\');
		if (slash > 0 && pref.GetValueBool("overall.task.autoCreateDir")) {
			if (MakeDir(instinfo.destfile.Left(slash).GetBuffer())) {
				str.Format(IDS_ERROR_MAKE_DIR, instinfo.destfile.GetBuffer());
				pfnEvent(MC_MSG_ALERT, str.GetBuffer());
				inst.state = STATE_ERROR;
				goto skip;
			}
		}
		} break;
	}

	{
	DWORD attr;
	str = pref.GetValueStr("overall.task.destdir");
	if (!str.IsEmpty()) {
		if (str.Right(1) == '\\') str.Delete(str.GetLength() - 1);
		attr = GetFileAttributes(str.GetBuffer());
		if (attr == INVALID_FILE_ATTRIBUTES || !(attr & FILE_ATTRIBUTE_DIRECTORY)) {
			str.Format(IDS_INVALID_DEST_DIR);
			pfnEvent(MC_MSG_ALERT, str.GetBuffer());
			goto skip;
		}
	}
	}

	inst.stats.dwClipStartTime = NOW;
	tm = inst.stats.dwClipStartTime - tm;
	inst.stats.dwEncStartTime += tm;
	inst.stats.dwTaskStartTime += tm;

	// encode audio
	if (mcinst.fAudioEncoding && (audioenc->flags & FLAG_NO_ENCODER)) {
		logger.Log(IDS_NOT_STANDALONE_ENCODER, 0);
		inst.state = STATE_ERROR;
	} else if (item->target.audioformat == AC_COPY) {
		mcContainerAddFile(CT_AUDIO, item->filename, DUMMY_FLAG);
	} else if (item->target.audioformat != AC_NONE && mcinst.fSoleAudio) {
		pfnEvent(MC_UPDATE_TITLE, "audio transcoding in progress");
		mcinst.audiochannel = pref.GetValueInt("overall.decoding.aid");
		for(;;) {
			if (StartTranscodeAudio(item, 0)) {
				if (!fAllTracks) {
					logger.Log(IDS_ERROR_AUDIO);
					inst.state = STATE_ERROR;
				}
				break;;
			}
			inst.stats.dwClipStartTime = GetTickCount();
			timerlevel = 2;
			WaitAudioEncoder(pref.GetValueInt("overall.task.timeout"));
            FinishTranscodeAudio(item);
			if (inst.state == STATE_ENCODING) {
				if (fAllTracks) {
					mcinst.audiochannel++;
					if (pref.GetValueInt("audiofilter.internal.downmix")) {
						if (mcinst.audiochannel == 2) break;
						inst.stats.dwClipStartTime = GetTickCount();
						switch (mcinst.audiochannel) {
						case 1:
							mcinst.audiofilter->pfnAudioDsp = dspRightChannelMono;
							break;
						}
					}
					continue;
				}
			}
			break;
		}
		if (pref.GetValueBool("overall.audio.allTracks")) pref.SetValue("overall.decoding.audioID", 0);
	} else if (!videoenc->CheckAudioSupport(item->target.audioformat)) {
		logger.Log(IDS_CANNOT_MUX_AUDIO);
	}
	pfnEvent(MC_UPDATE_STATS, 0);
	if (inst.state != STATE_ENCODING) goto skip;

	// encode video
	inst.stats.dwClipStartTime = GetTickCount();
	if (item->target.videoformat == VC_COPY && container->IsExternalMuxer()) {
		mcContainerAddFile(CT_VIDEO, item->filename, DUMMY_FLAG);
	} else if (!mcinst.fAudioEncoding) {
		if (TranscodeVideo(item)) {
			logger.Log(IDS_ERROR_VIDEO_TRANSCODE);
			inst.state=STATE_ERROR;
		}
	}
	mcinst.fNeedMux=false;
	if (inst.state!=STATE_ENCODING) goto skip;
	// mux into container
	inst.stage=ST_MUXING;
	if (container->IsExternalMuxer()) {
		mcPlayerStop();
		pfnEvent(MC_UPDATE_TITLE, "muxing tracks into container");
		pfnEvent(MC_UPDATE_STATE, (void*)inst.state);
		mcContainerMuxFiles(item, item->destfile);
	}
	skip:
	timerlevel = 0;
	// clean up
	mcContainerPurge(item, (pref.GetValueBool("overall.task.keepTemp") || item->target.container == CF_DEFAULT) ? 0 : 1);
	int destFileSize = GetFileKB(item->destfile);
	if (destFileSize > 0) {
		inst.stats.encodedKB += destFileSize;
	}
	if (inst.state != STATE_ENCODING && !pref.GetValueBool("overall.task.keepUnfinished")) {
		RemoveFile(item->destfile, 1000);
	}
	pfnEvent(MC_UPDATE_STATE, (void*)inst.state);

	// delete temporary files
	str.Format("%smcpass%d.log",pref.GetValueStr("overall.task.tempdir"),GetCurrentProcessId());
	DeleteFile(str.GetBuffer());
	str += ".temp";
	DeleteFile(str.GetBuffer());
	return destFileSize;
}

int mcTranscodeQueue(T_ITEM* queue)
{
	CString str;
	int errmsg = 0;
	T_ITEM* item;
	CLogger logger;

	if (!queue) queue = itemdef.next;
	if (!queue) return -1;

	// check files consistence in joining mode
	mcinst.taskMode= pref.GetValueInt("overall.task.mode");
	if (mcinst.taskMode == TM_JOIN) {
		mcGetMediaInfo(queue);
		if (pref.GetValueInt("overall.video.mode") >= 2 || pref.GetValueInt("overall.video.container") > 0 || pref.GetValueInt("overall.video.format") == 0) {
			str.Format(IDS_JOIN_TIP3);
			pfnEvent(MC_MSG_ALERT, str.GetBuffer());
			return -1;
		}
		if (!pref.GetValueInt("videofilter.scale.enabled")) {
			for (item = queue->next; item; item = item->next) {
				mcGetMediaInfo(item);
				if (item->info.video.width != queue->info.video.width || item->info.video.height != queue->info.video.height) {
					str.Format(IDS_JOIN_TIP1, queue->filename, item->filename);
					pfnEvent(MC_MSG_ALERT, str.GetBuffer());
					return -1;
				}
			}
		}
		if (!pref.GetValueInt("overall.video.fps")) {
			for (item = queue->next; item; item = item->next) {
				mcGetMediaInfo(item);
				if (item->info.video.fps != queue->info.video.fps) {
					str.Format(IDS_JOIN_TIP1, queue->filename, item->filename);
					pfnEvent(MC_MSG_ALERT, str.GetBuffer());
					return -1;
				}
			}
		}
		if (!pref.GetValueInt("audiofilter.resample.samplerate")) {
			for (item = queue->next; item; item = item->next) {
				mcGetMediaInfo(item);
				if (item->info.audio.samplerate != queue->info.audio.samplerate) {
					str.Format(IDS_JOIN_TIP2, queue->filename, item->filename);
					pfnEvent(MC_MSG_ALERT, str.GetBuffer());
					return -1;
				}
			}
		}
	}

	memset(&inst.stats, 0, sizeof(TASK_STATS));
	mcinst.fAudioDecodable = false;
	mcinst.ovaction = OA_ASK;
	mcinst.commonPartLen=0;
	if (pref.GetValueBool("overall.task.dupDir")) {
		char *p;
		item = itemdef.next;
		if (!strstr(item->filename, "://")) {
			p=strrchr(item->filename,'\\');
			if (p) {
				p++;
				mcinst.commonPartLen=p - item->filename;
			}
		}

		T_ITEM* prev = item;
		//find out the common part of paths
		for (item = itemdef.next; item && mcinst.commonPartLen>0; prev = item, item = item->next) {
			int i;
			if (strstr(item->filename, "://")) {
				mcinst.commonPartLen = 0;
				break;
			}
			for (i = ( p[1] ==':' ? 2 : 0); prev->filename[i] == item->filename[i] && i<mcinst.commonPartLen; i++);
			mcinst.commonPartLen=i;
		}
	}

	inst.state = STATE_ENCODING;
	inst.stats.dwEncStartTime = NOW;
	inst.stats.percent = 0;
	mcinst.startTimeOffset = 0;
	for (item = queue; item && inst.state == STATE_ENCODING; ) {
		mcFree((void**)&item->destfile);
		int destFileSize = mcTranscodeItem(item);
		if (mcinst.taskMode == TM_JOIN) break;
		if (mcinst.splitTime <= 0 || mcinst.startpos + mcinst.duration >= item->info.duration) {
			logger.Log(IDS_FINISH_JOB, item->filename);
			if (inst.state == STATE_ENCODING && destFileSize > 0) {
				int decodedMB = (unsigned int)(item->info.filesize >> 20);
				if (!decodedMB) decodedMB = 1;
				if (pref.GetValueBool("overall.task.deleteOriginal")) RemoveFile(item->filename, 3000);
				inst.stats.encodedFiles++;
				if (mcinst.duration && item->info.duration && (mcinst.duration < item->info.duration)) {
					inst.stats.decodedMB += decodedMB * (mcinst.duration / 1000) / (item->info.duration / 1000);
				} else {
					inst.stats.decodedMB += decodedMB;
				}
				item->flags |= IF_DONE;
			}
			pfnEvent(MC_JOB_FINISH, item);
			item = item->next;		// go on with next file
			mcinst.fileNameSuffix = 0;
			mcinst.startTimeOffset = 0;
		} else {
			mcinst.startTimeOffset += mcinst.splitTime;
			mcinst.fileNameSuffix++;
		}
		switch (inst.state) {
		case STATE_ERROR:
			inst.stats.errorFiles++;
			inst.state = STATE_ENCODING;
			break;
		case STATE_SKIPPING:
			inst.stats.skippedFiles++;
			inst.state = STATE_ENCODING;
			break;
		case STATE_STOPPING:
			inst.stats.skippedFiles++;
			break;
		case STATE_ENCODING: {
			// cooling down for a while
			int interval = pref.GetValueInt("overall.task.interval");
			pfnEvent(MC_UPDATE_TITLE, "cooling down...");
			for (int i = 0; i < interval && inst.state != STATE_STOPPING; i++) Sleep(1000);
			} break;
		}

	}

	mcApplyParam();
	inst.astats.percent = 100;
	pfnEvent(MC_UPDATE_TITLE, 0);
	pfnEvent(MC_UPDATE_PROGRESS, (void*)100);

	if (inst.state == STATE_STOPPING) {
		PopupMsg(IDS_TRANSCODE_INTERRUPTED);
		logger.Log(IDS_TRANSCODE_INTERRUPTED);
		inst.state = STATE_IDLE;
	} else if (!pref.GetValueBool("overall.task.shutdown")) {
		switch (inst.state) {
		case STATE_ENCODING:
			mcInvalidiateStatsData();
			if (!mcGetPrefValueInt("overall.task.shutdown") && inst.stats.encodedFiles > 0 && inst.stats.encodedKB > 0) {
				float ratio = inst.stats.encodedKB ? (float)inst.stats.decodedMB / inst.stats.encodedKB * 1024: 0;
				CString result;
				result.Format(IDS_TRANSCODE_DONE, inst.stats.encodedFiles, inst.stats.decodedMB, (float)inst.stats.encodedKB / 1024, (NOW - inst.stats.dwEncStartTime) / 1000, ratio);
				pfnEvent(MC_MSG_POPUP, result.GetBuffer());
			} else {
				PopupMsg(IDS_TRANSCODE_FAIL);
			}
			break;
		case STATE_ERROR:
			PopupMsg(IDS_TRANSCODE_FAIL);
			break;
		}
		inst.state = STATE_IDLE;
	} else {
		inst.state = STATE_SHUTDOWN;
	}
	SetEvent(mcinst.hEvent);
	mcinst.startpos = 0;
	mcinst.duration = 0;
	mcinst.startTimeOffset = 0;
	return 0;
}

STATES mcGetState()
{
	return inst.state;
}

void mcSetState(STATES state)
{
	inst.state = state;
}

void mcInvalidiateStatsData()
{
	inst.stats.dwElapsed = GetTickCount() - inst.stats.dwClipStartTime;
	if (inst.state != STATE_ENCODING) return;
	switch (inst.stage) {
	case ST_PROCESSING_AUDIO:
		if (audiosrc) {
			inst.astats.pos = audiosrc->GetPos();
			inst.astats.bytes = audiosrc->bytescount;
			if (inst.astats.pos > 0) {
				struct stat info;
				if (!stat(instinfo.audiofile.GetBuffer(), &info)) {
					inst.astats.bitrate = info.st_size * 8 / inst.astats.pos;
				}
				if (inst.astats.length > 0) {
					if (inst.astats.pos > inst.astats.length) {
						inst.astats.percent = 99;
					} else {
						inst.astats.percent = (float)inst.astats.pos * 100 / inst.astats.length;
						inst.astats.estsize = (int)((float)(info.st_size / 10) / inst.astats.percent);
					}
				}
			}
			if (inst.stats.dwElapsed) inst.astats.speed = inst.astats.pos * 10 / inst.stats.dwElapsed;
		}
		inst.stats.percent = inst.astats.percent;
		break;
	case ST_PROCESSING_VIDEO:
		inst.stats.percent = inst.vstats.percent;
		break;
	}
}

extern PARAM_DESC prefInfo[];

char* mcGetToolName(MC_TOOLS type, int nametype, int id)
{
	switch (type) {
	case TL_AUDIO_ENCODER:
		return nametype == 0 ? AudioEncoders[id]->GetName() : AudioEncoders[id]->GetFullName();
	case TL_VIDEO_CODEC:
		return pref.GetValueEnum("overall.video.format");
	case TL_VIDEO_ENCODER:
		return VideoEncoders[id]->GetName();
	case TL_CONTAINER:
		return Containers[id]->GetName();
	}
	return NULL;
}

char* mcGetToolVersion(MC_TOOLS type, int id)
{
	switch (type) {
	case TL_AUDIO_ENCODER:
		return AudioEncoders[id]->GetVersionInfo();
	case TL_VIDEO_ENCODER:
		return VideoEncoders[id]->GetVersionInfo();
	case TL_CONTAINER:
		return Containers[id]->GetVersionInfo();
	}
	return NULL;
}

char* mcGetToolCommand(MC_TOOLS type, int pass)
{
	int i;
	switch (type) {
	case TL_AUDIO_ENCODER:
		i = pref.GetValueInt("overall.audio.encoder");
		return AudioEncoders[i]->cl.GetBuffer();
	case TL_VIDEO_CODEC:
		if (pref.GetValueInt("overall.video.format") != VC_NONE) {
			i = pref.GetValueInt("overall.video.encoder");
			return VideoEncoders[i]->GetCommand(pass);
		}
		break;
	case TL_CONTAINER:
		//FIXME
		//return Containers[pref.GetValueInt("overall.video.container")]->cmd.GetBuffer();
		break;
	}
	return 0;	
}

char* mcRunMediaTool(MEDIA_TOOLS type, int action, T_ITEM* item)
{
	static CString result;
	int ret;
	CMediaTool *mt = (CMediaTool*)mcCreateObject(TL_TOOL, type);
	if (!mt) return 0;
	ret = mt->Run(action, item);
	result = mt->result;
    mcDestroyObject(TL_TOOL, type, mt);
	return ret ? 0 : result.GetBuffer();
}

extern HANDLE hEventSetupComplete;

char* mcShowConfig(CONFIG_DIALOG type, void* arg)
{
	int n;
	CString host;
	static CString url;
	host.Format("http://127.0.0.1:%d/", inst.httpdPort);
	url = host + "pref/?path=";

	switch (type) {
	case CD_PREF:
		if (arg) url += (char*)arg;
		url += "&cat=1";
		break;
	case CD_AUDIO_SRC:
		n = mcGetPrefValueInt("overall.audio.source");
		if (n == AS_WINAMP) {
			url = host + "pref/plugin.htm?type=input";
		} else {
			url += AudioSources[n]->prefNodePath;
		}
		break;
	case CD_AUDIO_ENC:
		url += AudioEncoders[pref.GetValueInt("overall.audio.encoder")]->prefNodePath;
		break;
	case CD_DSP_PLUGIN:
		url = host + "pref/plugin.htm?type=dsp";
		break;
	case CD_CONTAINER:
		url += Containers[pref.GetValueInt("overall.video.container")]->prefNodePath;
		break;
	case CD_SETUP:
		url = host + "setup/";
		if (!arg) {
			url += "setup1.htm";
		} else if (arg == (void*)1) {
			hEventSetupComplete = CreateEvent(0, 0, 0, 0);
		} else {
			WaitForSingleObject(hEventSetupComplete, 60000);
			CloseHandle(hEventSetupComplete);
			return 0;
		}
		break;
	default:
		return 0;
	}
	return url.GetBuffer();
}

void mcLogInfo(int id, char* arg)
{
	static CLogger applog;
	applog.Log(id, arg);
	pfnEvent(MC_UPDATE_LOG, 0);
}

#ifdef _SSE
#define VER_INFO "MediaCoder Core build %d (built on %s)"
#else
#define VER_INFO "MediaCoder Core (i686) build %d (built on %s)"
#endif

int mcGetBuildNo()
{
	return BUILD_NO;
}

char* mcGetVersionInfo()
{
	static char buf[64];
	_snprintf(buf, sizeof(buf), VER_INFO, BUILD_NO, BUILD_DATE);
	return buf;
}

int mcGetMediaInfo(T_ITEM* item)
{
	if (!item) return -1;

	if (itemdef.target.video.fps && !item->target.video.fps)
		item->target.video.fps = itemdef.target.video.fps;
	if (itemdef.target.video.width && !item->target.video.width)
		item->target.video.width = itemdef.target.video.width;
	if (itemdef.target.video.height && !item->target.video.height)
		item->target.video.height = itemdef.target.video.height;

	if (item->flags & IF_INFO_RETRIEVED) return 0;

	switch (item->info.filetype->exttype) {
	case FT_CDA:
		item->flags |= IF_INFO_RETRIEVED;
		item->info.audio.codec = "CDDA";
		item->info.audio.bitrate = 44100 * 4;
		item->info.audio.bitratemode = 1;
		item->info.audio.channels = 2;
		item->info.audio.samplebits = 16;
		item->info.audio.samplerate = 44100;
		return 0;
	}

	if (item->filename && FileExist(item->filename)) {
		if (!item->info.filesize)
			item->info.filesize = GetFileBytes(item->filename);
		if (!item->target.destsize)
			item->target.destsize = (int)(item->info.filesize / 2 / 1024);
		if (!item->target.ratio)
			item->target.ratio = 1;

		switch (item->info.filetype->exttype) {
		case FT_MP3:
		case FT_CUE:
		case FT_WAV:
		case FT_MPEG_TS:
		case FT_UNKNOWN:
			break;
		default: {
			CMediaInfo mi;
			mi.Run(1, item);
			if (item->info.duration) {
				if ((item->info.filetype->type == CT_AUDIO && (item->info.audio.codec || item->info.audio.samplerate)) || 
						(item->info.video.width && item->info.video.height)) {
					item->flags |= IF_INFO_RETRIEVED;
				}
			}
			} break;
		}
	}
	if (!(item->flags & IF_INFO_RETRIEVED)) {
		CMPParser mpp;
		mpp.Run(0, item);
		item->flags |= IF_INFO_RETRIEVED;
	}
	if (!item->target.crop.w) item->target.crop.w = item->info.video.width;
	if (!item->target.crop.h) item->target.crop.h = item->info.video.height;
	// retrieve file tags
	if (!item->info.tags.valid)
		ReadFileTags(item);
	return item->info.duration ? 0 : -1;
}

int mcGetTargetParam(TARGET_PARAM type)
{
	switch (type) {
	case TP_AUDIO_BITRATE:
		return AudioEncoders[itemdef.target.audioformat]->GetTargetBitRate(&itemdef);
	case TP_VIDEO_BITRATE:
        return VideoEncoders[itemdef.target.audioformat]->GetTargetBitRate(&itemdef);
	default:
		return -1;
	}
}

int mcSetTargetParam(TARGET_PARAM type, int val)
{
	return -1;
}

char* mcGetLastError()
{
	return mcinst.errmsg;
}

char* CDefContainer::GetOutExt(T_ITEM* item)
{
	return (mcinst.fAudioEncoding) ? audioenc->GetOutExt(item) : videoenc->GetOutExt(item);
}

char* CRawContainer::GetOutExt(T_ITEM* item)
{
	return (mcinst.fAudioEncoding) ? audioenc->GetOutExt(item) : videoenc->GetOutExt(item);
}
